;----------------------------------------------------------------------------------------------
; Start Here!
; This program drives my version of the Let's Make Robots "Start Here" project.  It is my first 
; completed PICMicro project and comes with no promises to be suitable for anything!
; - July 2010
;----------------------------------------------------------------------------------------------

list p=16F690
#include <p16F690.inc>
__config (_INTRC_OSC_NOCLKOUT & _WDT_OFF & _PWRTE_OFF & _MCLRE_OFF & _CP_OFF & _BOR_OFF & _IESO_OFF & _FCMEN_OFF)

;----------------------------------------------------------------------------------------------
;  Constants
;----------------------------------------------------------------------------------------------

ObstacleThreshold equ 0x50        ; how close an obstacle can be before an avoid sequence kicks in

PanRight          equ 0x33        ; position settings for the servos
PanCenter         equ 0x8C
PanLeft           equ 0xE5
TiltLo            equ 0xBF
TiltHi            equ 0x99

RightObstacleBit  equ 0x00        ; there is a byte variable Obstacle - individual bits will be used 
LeftObstacleBit   equ 0x01        ; to track if an obstacle has been detected left, right, or center
CenterObstacleBit equ 0x02

;----------------------------------------------------------------------------------------------
;  Variables
;----------------------------------------------------------------------------------------------

Context UDATA                    ; the context variables are used to save state during interupts
  w_temp       res 1
  status_temp  res 1
  pclath_temp  res 1

GenVar UDATA
  sPORTC           res 1         ; keeps a shadow of PORTC to avoid read/write/modify issues
  PulseCounter     res 1         ; used in timing the servo pulses
  Servo1Pos        res 1         ; hold the current servo position values
  Servo2Pos        res 1
  DelayCyclesOuter res 1         ; generic delay variables
  DelayCyclesInner res 1
  ChargeDelay      res 1         ; used in the ADC routine
  AnalogValue      res 1
  Obstacle         res 1         ; individual bits are used to track if an obstacle has been detected
  
;----------------------------------------------------------------------------------------------
; because this program uses an interupt which must start at 0x0004, 
; at 0x0000 there will only be a goto which jumps to a section of code labeled "Start"
;----------------------------------------------------------------------------------------------

INI CODE 0x0000
  pagesel Start
  goto    Start

;----------------------------------------------------------------------------------------------
;  Interupt Service Routine
;  will execute to refresh the servos with each TMRO overflow  the timer has been 
;  configured in the Start section below to interupt every 16ms
;----------------------------------------------------------------------------------------------

ISR CODE 0x004

;----------------------------------------------------------------------------------------------
; save context - stores the current values of W, STATUS and PCLATH - they will be restored at
; end of the interupt routine
;----------------------------------------------------------------------------------------------

  movwf   w_temp          
  movf    STATUS,w 
  movwf   status_temp
  movf    PCLATH,w
  movwf   pclath_temp

;----------------------------------------------------------------------------------------------
; execute code
; servos need a pulse of between 1-2ms - to set the angle these routines bring the pin driving
; the servo high using a delay of 1ms and then initiate another delay that varies from 0-1ms
; when the delays are completed the pin is brought low again
;----------------------------------------------------------------------------------------------
  banksel PORTC
  movlw   b'00100000'  ; turn on RC5 (the PAN servo) by flipping bit 5 in PORTC
  xorwf   sPORTC,1     ; using a shadow port to hold the values and then copying the shadow into
  movf    sPORTC,w     ; the actual port solves problems encountered when trying to read port pins
  movwf   PORTC        ; to determine whether they are on or off  
  movlw   0xFF         ; load the PulseCounter variable with 0xFF = ~1ms
  movwf   PulseCounter 
  call    Pulse
  movfw   Servo1Pos    ; load the PulseCounter variable with the value
  movwf   PulseCounter ; to add to 1ms to provide the proper angle of the servo
  call    Pulse        
  movlw   b'00001111'  ; turn off RC5
  andwf   sPORTC,1
  movf    sPORTC,w
  movwf   PORTC

  movlw   b'00010000'  ; turn on RC4 (the TILT servo)
  xorwf   sPORTC,1
  movf    sPORTC,w
  movwf   PORTC
  movlw   0xFF
  movwf   PulseCounter
  call    Pulse
  movfw   Servo2Pos
  movwf   PulseCounter
  call    Pulse
  movlw   b'00001111'  ; turn off RC4
  andwf   sPORTC,1
  movf    sPORTC,w
  movwf   PORTC
;----------------------------------------------------------------------------------------------
; restore context
;----------------------------------------------------------------------------------------------
  banksel PCLATH
  movf    pclath_temp,W   
  movwf   PCLATH
  banksel STATUS
  movf    status_temp,W   
  movwf   STATUS          
  swapf   w_temp,f       ; these swap routines (rather than mov) preserve the Z flag
  swapf   w_temp,w 
  bcf     INTCON, T0IF   ; clear the overflow flag
  retfie

MAIN CODE

;----------------------------------------------------------------------------------------------
; Initialize the registers
;     - PORTA = input | PORTC = output
;     - RC0 to RC3 will drive the motors through an H-bridge
;     - RC4 and RC5 will be servo outputs
;     - make PORTA analog and configure RA1 for ADC input
;     - configure the timer
;     - enable the timer and global interupts
;----------------------------------------------------------------------------------------------

Start:
     banksel   TRISA
     movlw     0xFF
     movwf     TRISA          ; Make PortA all input
     clrf      TRISC          ; Make PortC all output
     movlw     b'10000101'    ; configure Prescaler on Timer0, prescale 101 = 1:64 = ~16ms
     movwf     OPTION_REG     ; bit 7 = pull-up resistors bit 2:0 = prescaler options
     movlw     b'10100000'    ; enable Timer 0 and global interrupts
     movwf     INTCON         ; bit 7 = GIE bit 5 = T0IE
     banksel   ANSEL
     movlw     0xFF 
     movwf     ANSEL          ; sets all PortA pins Analog
     banksel   ADCON0
     movlw     b'00000101'
     movwf     ADCON0         ; configure A2D for Channel 1 (RA1), Left justified, and turn on the A2D module
     banksel   PORTC

;----------------------------------------------------------------------------------------------
; the main routine will
;    - move the servos to a sequence of six pan/tilt positions,
;    - take a reading and store the result in the appropriate variable
;    - determine if there are any obtstacles
;    - issue the appropriate command to the motors
;----------------------------------------------------------------------------------------------

Main:

   movlw     0x00
   movwf     Obstacle

   movlw     PanRight                  ; load the servo positioning variables with predefined constants
   movwf     Servo1Pos                 
   movlw     TiltLo
   movwf     Servo2Pos

   movlw     0xFE                      ; put 0xFE into working register
   movwf     DelayCyclesOuter          ; load DelayCyclesOuter with that value
   call      Delay                     ; execute the Delay subroutine
   
   call      ADCLoop                   ; take a sensor reading
   movfw     AnalogValue               ; load that reading into the W register
   sublw     ObstacleThreshold         ; subtracts ObstacleThreshold from W and stores result in W
   btfss     STATUS,0                  
;----------------------------------------------------------------------------------------------
; btfss = "bit test f skip if set" 
; bit 0 of the STATUS register is the carry/borrow bit - btfss tests to see if the 
; sublw subtraction resulted in a borrow - if yes it means that the threshold is greater than   
; the sensor reading and therefore skips the next instruction which would set the bit for an 
; obstacle being detected - if the sensor reading is greater than the threshold then the borrow
; will not occur, the next line will not be skipped and the obstacle bit will be set
;----------------------------------------------------------------------------------------------
   bsf       Obstacle,RightObstacleBit 
                                       
                                       
 
   movlw     PanCenter
   movwf     Servo1Pos
   movlw     TiltLo
   movwf     Servo2Pos

   movlw     0xFE              
   movwf     DelayCyclesOuter    
   call      Delay
   
   call      ADCLoop
   movfw     AnalogValue
   sublw     ObstacleThreshold   
   btfss     STATUS,0            
   bsf       Obstacle,CenterObstacleBit        

   movlw     PanLeft
   movwf     Servo1Pos
   movlw     TiltLo
   movwf     Servo2Pos

   movlw     0xFE               
   movwf     DelayCyclesOuter    
   call      Delay
   
   call      ADCLoop
   movfw     AnalogValue
   sublw     ObstacleThreshold   
   btfss     STATUS,0            
   bsf       Obstacle,LeftObstacleBit          

   movlw     PanLeft
   movwf     Servo1Pos
   movlw     TiltHi
   movwf     Servo2Pos

   movlw     0xFE               
   movwf     DelayCyclesOuter    
   call      Delay
   
   call      ADCLoop
   movfw     AnalogValue
   sublw     ObstacleThreshold   
   btfss     STATUS,0            
   bsf       Obstacle,LeftObstacleBit           


   movlw     PanCenter
   movwf     Servo1Pos
   movlw     TiltHi
   movwf     Servo2Pos

   movlw     0xFE 
   movwf     DelayCyclesOuter  
   call      Delay
   
   call      ADCLoop
   movfw     AnalogValue
   sublw     ObstacleThreshold   
   btfss     STATUS,0            
   bsf       Obstacle,CenterObstacleBit          

   movlw     PanRight
   movwf     Servo1Pos
   movlw     TiltHi
   movwf     Servo2Pos

   movlw     0xFE               
   movwf     DelayCyclesOuter   
   call      Delay
   
   call      ADCLoop
   movfw     AnalogValue
   sublw     ObstacleThreshold   
   btfss     STATUS,0            
   bsf       Obstacle,RightObstacleBit

;----------------------------------------------------------------------------------------------
; test each obstacle bit in succession - if a bit is set then execute the
; avoidance and return to main with the motors off - if all bits are clear then
; set the motors going forward and continue cycling the servo positions and testing 
; for obstacles
;----------------------------------------------------------------------------------------------
   

   btfss     Obstacle,RightObstacleBit     ; bit test f skip if set - if the RightObstactleBit is not detected
   goto      LeftTest                      ; then the code branches to the next test
   call      StopWheels                    ; if the obstacle bit was set the avoid  
   call      QuarterLeft                   ; routine executes and 
   goto      Main                          ; then the code branches to Main
LeftTest:
   btfss     Obstacle,LeftObstacleBit     
   goto      CenterTest
   call      StopWheels
   call      QuarterRight
   goto      Main
CenterTest:
   btfss     Obstacle,CenterObstacleBit     
   goto      AllClear
   call      StopWheels
   call      QuarterLeft
   goto      Main
AllClear:
   call      ContinuousForward
   goto      Main

;----------------------------------------------------------------------------------------------
; take an analog reading of pin RA0
; store the result in AnalogValue
;----------------------------------------------------------------------------------------------

ADCLoop:
   movlw     0x07                ; put 0x07 into working register
   movwf     ChargeDelay         ; load ChargeDelay with that value
ChargeLoop:                      ; ChargeLoop should create a delay of ~22us to allow the ADC capacitor to charge
   decfsz    ChargeDelay,1
   goto      ChargeLoop     
   bsf       ADCON0,GO          ; start conversion
   btfss     ADCON0,GO          ; this bit will change to zero when the conversion is complete
   goto      $-1
   movf      ADRESH,0           ; loads the high byte of the analog value to W
   movwf     AnalogValue        ; and then into the variable AnalogValue
   return

;----------------------------------------------------------------------------------------------
; locomotion subroutines
; - these routines simply flip the pins driving an h-bridge motor controller
;----------------------------------------------------------------------------------------------

ContinuousForward:
  movlw      b'11110000'
  andwf      sPORTC,1
  movf       sPORTC,w
  movwf      PORTC
  movlw      b'00001001'
  xorwf      sPORTC,1
  movf       sPORTC,w
  movwf      PORTC
  return


StopWheels:

  movlw      b'11110000'
  andwf      sPORTC,1
  movf       sPORTC,w
  movwf      PORTC
  return

QuarterLeft:

  movlw      b'11110000'
  andwf      sPORTC,1
  movf       sPORTC,w
  movwf      PORTC
  movlw      b'00000101'
  xorwf      sPORTC,1
  movf       sPORTC,w
  movwf      PORTC

  movlw      0xFE                ; put 0xFE into working register
  movwf      DelayCyclesOuter    ; load DelayCyclesOuter with that value
  call       Delay
  movlw      0xFE                
  movwf      DelayCyclesOuter 
  call       Delay

  movlw      b'11110000'
  andwf      sPORTC,1
  movf       sPORTC,w
  movwf      PORTC

  return

QuarterRight:

  movlw      b'11110000'
  andwf      sPORTC,1
  movf       sPORTC,w
  movwf      PORTC
  movlw      b'00001010'
  xorwf      sPORTC,1
  movf       sPORTC,w
  movwf      PORTC

  movlw      0xFE               
  movwf      DelayCyclesOuter   
  call       Delay
  movlw      0xFE             
  movwf      DelayCyclesOuter 
  call       Delay

  movlw      b'11110000'
  andwf      sPORTC,1
  movf       sPORTC,w
  movwf      PORTC

  return

;----------------------------------------------------------------------------------------------
; Delay loops are built based on the notion of wasting time.  A sequence of nested loops that
; have no inherent function executes.  By calculating how many instruction cycles are consumed,
; the amount of time passed is determined.  One instruction cycle is equal to four clock cycles.  
; On a chip running at 4 mhz, instructions are executed at 1 mhz or 1us per instruction.
; Most instructions only take one cycle.  GOTO, CALL, and the various returns take 2 cycles.
; The conditional instructions take 1 cycle if false, 2 cycles if true.
;----------------------------------------------------------------------------------------------


;----------------------------------------------------------------------------------------------
; process pulse delays
;----------------------------------------------------------------------------------------------

Pulse:
   nop
   decfsz    PulseCounter,1
   goto      Pulse
   return

;----------------------------------------------------------------------------------------------
; generic delay
;----------------------------------------------------------------------------------------------

Delay:
   clrf      DelayCyclesInner   ; set DelayCyclesInner to 0 to create an inner loop of 256

DlyLoops:
   nop
   decfsz    DelayCyclesInner,1
   goto      DlyLoops
   decfsz    DelayCyclesOuter,1
   goto      DlyLoops    

   return

end